{***************************************************************************}
{                        UNIDAD DE FRASES COMO NUMEROS

                                  Copyright (c) 1998 Marco Antonio Alvarado
}
{***************************************************************************}

UNIT UFrasNum;

INTERFACE

  FUNCTION AjustaTamanoCeros (A : STRING; Tamano : Byte) : STRING;
  FUNCTION QuitaCerosIzquierda (A : STRING) : STRING;

{***************************************************************************}

  FUNCTION ValorCifra (Cifra : Char) : Byte;
  FUNCTION CifraValor (Valor : Byte) : Char;
  FUNCTION ValorSigno (A : STRING) : ShortInt;
  FUNCTION Niega (A : STRING) : STRING;
  FUNCTION Absoluto (A : STRING) : STRING;
  FUNCTION Igual (A, B : STRING) : Boolean;
  FUNCTION Mayor (A, B : STRING) : Boolean;
  FUNCTION MayorIgual (A, B : STRING) : Boolean;
  FUNCTION EnRango (A, MenorB, MayorB : STRING) : Boolean;
  FUNCTION Suma (A, B : STRING) : STRING;
  FUNCTION Resta (A, B : STRING) : STRING;
  FUNCTION Multiplica (A, B : STRING) : STRING;
  FUNCTION Divide (A, B : STRING; VAR Residuo : STRING) : STRING;

{***************************************************************************}

  FUNCTION FraseValor (Valor : LongInt) : STRING;
  FUNCTION ValorFrase (Frase : STRING) : LongInt;
  FUNCTION Convierte (A : STRING; Base : Byte) : STRING;
  FUNCTION ConvierteBase (A : STRING; Base : Byte) : STRING;

IMPLEMENTATION

  USES
    UFrase;

  CONST
    MayorCifra = 35;
    Cifras : ARRAY [0..MayorCifra] OF Char =
      '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  FUNCTION AjustaTamanoCeros (A : STRING; Tamano : Byte) : STRING;

    VAR
      TamanoA : Byte ABSOLUTE A;

  BEGIN

    IF Tamano > TamanoA THEN
    BEGIN
      Dec (Tamano, TamanoA);

      IF A [1] = '-' THEN
      BEGIN
{Conserva el signo.}
        A := Copy (A, 2, TamanoA - 1);
        FOR Tamano := Tamano DOWNTO 1 DO A := '0' + A;
        A := '-' + A;
      END ELSE
      BEGIN
        FOR Tamano := Tamano DOWNTO 1 DO A := '0' + A;
      END;

    END;

    AjustaTamanoCeros := A;
  END;

  FUNCTION QuitaCerosIzquierda (A : STRING) : STRING;

    VAR
      Tamano : Byte ABSOLUTE A;
      Indice : Byte;

  BEGIN

    IF A [1] = '-' THEN
    BEGIN
{Conserva el signo.}
      Indice := 2;
      WHILE (A [Indice] = '0') AND (Indice < Tamano) DO Inc (Indice);
      QuitaCerosIzquierda := '-' + Copy (A, Indice, Tamano - Indice + 1);
    END ELSE
    BEGIN
      Indice := 1;
      WHILE (A [Indice] = '0') AND (Indice < Tamano) DO Inc (Indice);
      QuitaCerosIzquierda := Copy (A, Indice, Tamano - Indice + 1);
    END;

  END;

{***************************************************************************}

  FUNCTION ValorCifra (Cifra : Char) : Byte;

    VAR
      Indice : Byte;

  BEGIN
    Cifra := SubeLetra (Cifra);
    Indice := MayorCifra;
    WHILE (Indice > 0) AND (Cifra <> Cifras [Indice]) DO Dec (Indice);
{Las cifras desconocidas valen 0.}
    ValorCifra := Indice;
  END;

  FUNCTION CifraValor (Valor : Byte) : Char;
  BEGIN
    IF Valor <= MayorCifra THEN CifraValor := Cifras [Valor] ELSE
      CifraValor := '0';
  END;

  FUNCTION ValorSigno (A : STRING) : ShortInt;
  BEGIN
    A := QuitaCerosIzquierda (A);
    IF A [1] = '-' THEN ValorSigno := -1 ELSE ValorSigno := 1;
  END;

  FUNCTION Niega (A : STRING) : STRING;

    VAR
      Tamano : Byte ABSOLUTE A;

  BEGIN

    IF A [1] = '-' THEN
    BEGIN
{Quita el signo.}
      Niega := Copy (A, 2, Tamano - 1);
    END ELSE
    BEGIN
{Agrega el signo.}
      Niega := '-' + A;
    END;

  END;

  FUNCTION Absoluto (A : STRING) : STRING;
  BEGIN
    IF ValorSigno (A) = 1 THEN Absoluto := A ELSE Absoluto := Niega (A);
  END;

  FUNCTION Suma (A, B : STRING) : STRING;

    VAR
      TamanoA : Byte ABSOLUTE A;
      TamanoB : Byte ABSOLUTE B;
      SignoA,
      SignoB : ShortInt;
      Indice : Byte;
      Valor : Integer;
      Resultado : STRING;
      Acarreo : ShortInt;

  BEGIN
    SignoA := ValorSigno (A);
    SignoB := ValorSigno (B);
    A := Absoluto (QuitaCerosIzquierda (A));
    B := Absoluto (QuitaCerosIzquierda (B));
    IF TamanoA < TamanoB THEN A := AjustaTamanoCeros (A, TamanoB) ELSE
    IF TamanoB < TamanoA THEN B := AjustaTamanoCeros (B, TamanoA);
    Resultado := '';

    IF (A = '0') AND (B = '0') THEN Resultado := '0' ELSE
    IF SignoA = SignoB THEN
    BEGIN
      Acarreo := 0;

      FOR Indice := TamanoA DOWNTO 1 DO
      BEGIN
        Valor := Acarreo + ValorCifra (A [Indice]) + ValorCifra (B [Indice]);

        IF Valor < 10 THEN Acarreo := 0 ELSE
        BEGIN
          Acarreo := 1;
          Dec (Valor, 10);
        END;

        Resultado := CifraValor (Valor) + Resultado;
      END;

      IF Acarreo = 1 THEN Resultado := '1' + Resultado;
      Resultado := QuitaCerosIzquierda (Resultado);
      IF SignoA = -1 THEN Resultado := '-' + Resultado;
    END ELSE
    BEGIN

      IF Igual (A, B) THEN Resultado := '0' ELSE
      IF Mayor (A, B) THEN
      BEGIN
        Acarreo := 0;

        FOR Indice := TamanoA DOWNTO 1 DO
        BEGIN
          Valor := Acarreo + ValorCifra (A [Indice]) - ValorCifra (B
            [Indice]);

          IF Valor < 0 THEN
          BEGIN
            Acarreo := -1;
            Inc (Valor, 10);
          END ELSE Acarreo := 0;

          Resultado := CifraValor (Valor) + Resultado;
        END;

        Resultado := QuitaCerosIzquierda (Resultado);
        IF SignoA = -1 THEN Resultado := '-' + Resultado;
      END ELSE
      BEGIN
        Acarreo := 0;

        FOR Indice := TamanoA DOWNTO 1 DO
        BEGIN
          Valor := Acarreo + ValorCifra (B [Indice]) - ValorCifra (A [Indice]);

          IF Valor < 0 THEN
          BEGIN
            Acarreo := -1;
            Inc (Valor, 10);
          END ELSE Acarreo := 0;

          Resultado := CifraValor (Valor) + Resultado;
        END;

        Resultado := QuitaCerosIzquierda (Resultado);
        IF SignoB = -1 THEN Resultado := '-' + Resultado;
      END;

    END;

    Suma := Resultado;
  END;

  FUNCTION Igual (A, B : STRING) : Boolean;
  BEGIN
    Igual := QuitaCerosIzquierda (A) = QuitaCerosIzquierda (B);
  END;

  FUNCTION Mayor (A, B : STRING) : Boolean;

    VAR
      TamanoA : Byte ABSOLUTE A;
      TamanoB : Byte ABSOLUTE B;
      SignoA,
      SignoB : ShortInt;
      Indice : Byte;

  BEGIN
    SignoA := ValorSigno (A);
    SignoB := ValorSigno (B);

    IF SignoA > SignoB THEN Mayor := True ELSE
    IF SignoA < SignoB THEN Mayor := False ELSE
    BEGIN
      A := Absoluto (QuitaCerosIzquierda (A));
      B := Absoluto (QuitaCerosIzquierda (B));

      IF SignoA > 0 THEN
      BEGIN

        IF TamanoA > TamanoB THEN Mayor := True ELSE
        IF TamanoA < TamanoB THEN Mayor := False ELSE
        BEGIN
          Indice := 0;

          REPEAT
            Inc (Indice);
            IF ValorCifra (A [Indice]) > ValorCifra (B [Indice]) THEN
              Mayor := True ELSE Mayor := False;
          UNTIL (Indice = TamanoA) OR (A [Indice] <> B [Indice]);

        END;

      END ELSE
      BEGIN

        IF TamanoA > TamanoB THEN Mayor := False ELSE
        IF TamanoA < TamanoB THEN Mayor := True ELSE
        BEGIN
          Indice := 0;

          REPEAT
            Inc (Indice);
            IF ValorCifra (A [Indice]) < ValorCifra (B [Indice]) THEN
              Mayor := True ELSE Mayor := False;
          UNTIL (Indice = TamanoA) OR (A [Indice] <> B [Indice]);

        END;

      END;

    END;

  END;

  FUNCTION MayorIgual (A, B : STRING) : Boolean;

    VAR
      TamanoA : Byte ABSOLUTE A;
      TamanoB : Byte ABSOLUTE B;
      SignoA,
      SignoB : ShortInt;
      Indice : Byte;

  BEGIN
    A := QuitaCerosIzquierda (A);
    B := QuitaCerosIzquierda (B);

    IF A = B THEN MayorIgual := True ELSE
    BEGIN
      SignoA := ValorSigno (A);
      SignoB := ValorSigno (B);

      IF SignoA > SignoB THEN MayorIgual := True ELSE
      IF SignoA < SignoB THEN MayorIgual := False ELSE
      BEGIN
        A := Absoluto (A);
        B := Absoluto (B);

        IF SignoA > 0 THEN
        BEGIN

          IF TamanoA > TamanoB THEN MayorIgual := True ELSE
          IF TamanoA < TamanoB THEN MayorIgual := False ELSE
          BEGIN
            Indice := 0;

            REPEAT
              Inc (Indice);
              IF ValorCifra (A [Indice]) > ValorCifra (B [Indice]) THEN
                MayorIgual := True ELSE MayorIgual := False;
            UNTIL (Indice = TamanoA) OR (A [Indice] <> B [Indice]);

          END;

        END ELSE
        BEGIN

          IF TamanoA > TamanoB THEN MayorIgual := False ELSE
          IF TamanoA < TamanoB THEN MayorIgual := True ELSE
          BEGIN
            Indice := 0;

            REPEAT
              Inc (Indice);
              IF ValorCifra (A [Indice]) < ValorCifra (B [Indice]) THEN
                MayorIgual := True ELSE MayorIgual := False;
            UNTIL (Indice = TamanoA) OR (A [Indice] <> B [Indice]);

          END;

        END;

      END;

    END;

  END;

  FUNCTION EnRango (A, MenorB, MayorB : STRING) : Boolean;
  BEGIN
    EnRango := (MayorIgual (A, MenorB)) AND (MayorIgual (MayorB, A));
  END;

  FUNCTION Resta (A, B : STRING) : STRING;
  BEGIN
    Resta := Suma (A, Niega (B));
  END;

  FUNCTION Multiplica (A, B : STRING) : STRING;

    VAR
      TamanoA : Byte ABSOLUTE A;
      TamanoB : Byte ABSOLUTE B;
      Signo : ShortInt;
      Resultado,
      Resultado1 : STRING;
      IndiceB,
      IndiceA,
      Acarreo,
      Factor,
      Valor,
      IndicePosicion,
      Posicion : Byte;

  BEGIN
    Signo := ValorSigno (A) * ValorSigno (B);
    A := Absoluto (QuitaCerosIzquierda (A));
    B := Absoluto (QuitaCerosIzquierda (B));
    Resultado := '0';

    IF (A <> '0') AND (B <> '0') THEN
    IF A = '1' THEN
    BEGIN
      Resultado := B;
    END ELSE
    IF B = '1' THEN
    BEGIN
      Resultado := A;
    END ELSE
    BEGIN
      Posicion := 0;

      FOR IndiceB := TamanoB DOWNTO 1 DO
      BEGIN
        Acarreo := 0;
        Factor := ValorCifra (B [IndiceB]);
        Resultado1 := '';

        FOR IndiceA := TamanoA DOWNTO 1 DO
        BEGIN
          Valor := Acarreo + (Factor * ValorCifra (A [IndiceA]));

          IF Valor < 10 THEN Acarreo := 0 ELSE
          BEGIN
            Acarreo := Valor DIV 10;
            Valor := Valor MOD 10;
          END;

          Resultado1 := CifraValor (Valor) + Resultado1;
        END;

        IF Acarreo > 0 THEN Resultado1 := CifraValor (Acarreo) + Resultado1;

        FOR IndicePosicion := 1 TO Posicion DO
          Resultado1 := Resultado1 + '0';
        Inc (Posicion);
        Resultado := Suma (Resultado, Resultado1);
      END;

      Resultado := QuitaCerosIzquierda (Resultado);
      IF Signo = -1 THEN Resultado := '-' + Resultado;
    END;

    Multiplica := Resultado;
  END;

  FUNCTION Divide (A, B : STRING; VAR Residuo : STRING) : STRING;

    VAR
      TamanoA : Byte ABSOLUTE A;
      Signo : ShortInt;
      IndiceA : Byte;
      ParteA,
      Resultado,
      Substraccion : STRING;
      Valor : Byte;

  BEGIN

    IF B = '0' THEN
    BEGIN
{El resultado es indeterminable para la divisin entre cero.}
      Resultado := '';
      Residuo := '';
    END ELSE
    BEGIN
      Signo := ValorSigno (A) * ValorSigno (B);
      A := Absoluto (QuitaCerosIzquierda (A));
      B := Absoluto (QuitaCerosIzquierda (B));

      IF Igual (A, B) THEN
      BEGIN
        Resultado := '1';
        Residuo := '0';
      END ELSE
      IF A = '0' THEN
      BEGIN
        Resultado := '0';
        Residuo := '0';
      END ELSE
      IF B = '1' THEN
      BEGIN
        Resultado := A;
        Residuo := '0';
      END ELSE
      IF Mayor (B, A) THEN
      BEGIN
        Resultado := '0';
        Residuo := A;
      END ELSE
      BEGIN
        IndiceA := 1;
        ParteA := A [IndiceA];
        Resultado := '';

        REPEAT

          IF (Mayor (B, ParteA)) AND (IndiceA < TamanoA) THEN
          BEGIN
            Inc (IndiceA);
            ParteA := ParteA + A [IndiceA];
          END;

          IF Mayor (B, ParteA) THEN Resultado := Resultado + '0' ELSE
          BEGIN
            Valor := 1;
            Substraccion := B;

            WHILE MayorIgual (ParteA, Suma (Substraccion, B)) DO
            BEGIN
              Inc (Valor);
              Substraccion := Suma (Substraccion, B);
            END;

            Resultado := Resultado + CifraValor (Valor);
            ParteA := Resta (ParteA, Substraccion);
          END;

        UNTIL (IndiceA = TamanoA) AND (Mayor (B, ParteA));

        Residuo := QuitaCerosIzquierda (ParteA);
      END;

      IF Signo = -1 THEN Resultado := '-' + Resultado;
    END;

    Divide := Resultado;
  END;

{***************************************************************************}

  FUNCTION FraseValor (Valor : LongInt) : STRING;

    VAR
      Signo : ShortInt;
      Resultado : STRING;

  BEGIN

    IF Valor = 0 THEN Resultado := '0' ELSE
    BEGIN

      IF Valor > 0 THEN Signo := 1 ELSE
      BEGIN
        Signo := -1;
        Valor := Abs (Valor);
      END;

      Resultado := '';

      WHILE Valor > 0 DO
      BEGIN
        Resultado := Cifras [Valor MOD 10] + Resultado;
        Valor := Valor DIV 10;
      END;

      IF Signo = -1 THEN Resultado := '-' + Resultado;
    END;

    FraseValor := Resultado;
  END;

  FUNCTION ValorFrase (Frase : STRING) : LongInt;

    VAR
      Signo : ShortInt;
      Resultado,
      Posicion : LongInt;
      Indice,
      IndicePosicion : Byte;
      Tamano : Byte ABSOLUTE Frase;

  BEGIN
    Signo := ValorSigno (Frase);
    IF Signo = -1 THEN Frase := Absoluto (QuitaCerosIzquierda (Frase)) ELSE
      Frase := QuitaCerosIzquierda (Frase);
    Resultado := 0;

    FOR Indice := 1 TO Tamano DO
    BEGIN
      Posicion := 1;
      FOR IndicePosicion := 2 TO Tamano - Indice + 1 DO
        Posicion := Posicion * 10;
      Inc (Resultado, ValorCifra (Frase [Indice]) * Posicion);
    END;

    ValorFrase := Resultado * Signo;
  END;

  FUNCTION Convierte (A : STRING; Base : Byte) : STRING;

    VAR
      TamanoA : Byte ABSOLUTE A;
      Signo : ShortInt;
      Indice : Byte;
      B,
      Posicion,
      Resultado : STRING;
      IndicePosicion : Byte;

  BEGIN
    Signo := ValorSigno (A);
    A := Absoluto (QuitaCerosIzquierda (A));
    B := FraseValor (Base);
    Resultado := '0';

    IF A <> '0' THEN
    BEGIN

      FOR Indice := 1 TO TamanoA DO
      BEGIN
        Posicion := '1';
        FOR IndicePosicion := 2 TO TamanoA - Indice + 1 DO
          Posicion := Multiplica (Posicion, B);
        Resultado := Suma (Resultado,
          Multiplica (FraseValor (ValorCifra (A [Indice])), Posicion));
      END;

      IF Signo = -1 THEN Resultado := '-' + Resultado;
    END;

    Convierte := Resultado;
  END;

  FUNCTION ConvierteBase (A : STRING; Base : Byte) : STRING;

    VAR
      Signo : ShortInt;
      B,
      Resultado,
      Residuo : STRING;

  BEGIN
    Signo := ValorSigno (A);
    A := Absoluto (QuitaCerosIzquierda (A));

    IF A = '0' THEN Resultado := '0' ELSE
    BEGIN
      B := FraseValor (Base);
      Resultado := '';

      WHILE Mayor (A, '0') DO
      BEGIN
        A := Divide (A, B, Residuo);
        Resultado := Cifras [ValorFrase (Residuo)] + Resultado;
      END;

      IF Signo = -1 THEN Resultado := '-' + Resultado;
    END;

    ConvierteBase := Resultado;
  END;

END.